# Chat

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# TODO: (njh)
#   - clickable urls in chat window(James Mollison)
#   - ability to remove people from click-list(Michale Kaufman)

# TODO: detect when someone drops off the network badly

# TODO: float pickle

# TODO: improve /recall : select from list, only recall from some people for single message

# TODO: offline message polling, in case of net split

# All messages in the chat widget are displayed in "fields",
# eg a field may display a message indicating some task is in progress, 
# then later change to say that it has finished

import os,os.path,string,re,cStringIO,traceback,select

from types import *

from settings import *
from utility import *
from error import Error
import chat_commands, safe_pickle, hash

def english_list(list):
    if len(list) == 1:
        return list[0]
    else:
        return string.join(list[:-1],', ')+' and '+list[-1]

def chat_tell_task(chat,dests,message, field, attachment=None):
    do_gtk(field.show,_('Sending message...\n'))

    result = ''
    try:
        dest_list = [ ]
        dest_address  = [ ]
        dest_names = [ ]
        dest_bad_names = [ ]
        offline_list  = [ ]
        offline_names = [ ]
        recipient_list = [ ]
        channel_pipes = { }
        for dest in dests:
            if dest[0]=='#':
                channel_pipes[dest] = chat.app.channels.sub_list_pipe(dest)
                recipient_list.append((None,None,dest))
                continue
            acq = chat.app.name_server.locate(dest)
            acq.start_watching(chat.node,1)

            acq.lock.acquire()
            online = acq.online
            address = acq.address
            name = acq.name
            username = acq.info['name']
            acq.lock.release()

            if not online:
                offline_list.append(acq)
                offline_names.append(dest)
                recipient_list.append((name,None,username))
                #result = result+dest+ ' is not connected.\n'
            else:
                dest_list.append(acq)
                dest_address.append(address)
                dest_names.append(dest)
                recipient_list.append((name,address,username))
        
        #if result:
        #  do_gtk(field.show,result)

        my_details = (
            chat.app.name_server.public_key_name,
            chat.app.name_server.name
        )
        if attachment == None:
            sig = chat.app.name_server.sign(message)
            message = ('chat message',my_details,recipient_list,message,sig)
        else:
            sig = chat.app.name_server.sign(message+attachment)
            message = ('chat message',my_details,recipient_list,message,attachment,sig)

        for i in range(len(dest_list)):
            try:
                ret_value = chat.node.call(dest_address[i],message)
                
                #result = result + 'Message sent to '+ dest_names[i] +'.\n'
                
                #Type checking?
                if type(ret_value) <> TupleType:
                    raise Error(_('unexpected value returned from call'))
                
                if ret_value[0]:
                    if ret_value[1] <> '':
                        result = result + dest_names[i] + ' ' + ret_value[1] + '\n'
                else:
                    if ret_value[1] == '':
                        result = result + dest_names[i] + _(' is quiet.\n')
                    else:
                        result = result + dest_names[i] + _(' is quiet: ') + ret_value[1] + '\n'
            except Error:
                offline_list.append(dest_list[i])
                offline_names.append(dest_names[i])
                dest_bad_names.append(dest_names[i])
                result = result + dest_names[i] + _(' could not be contacted.\n')

        for name in dest_bad_names:
            dest_names.remove(name)

        #prev_result = result
        if dest_names:
            result = _('Message sent to %s.')%english_list(dest_names) + '\n' + result

        do_gtk(field.show,result+_('Sending message...\n'))
            
        #for pipe in channel_pipe_list:
        #  chcount=0

        channel_counts = { }
        for key in channel_pipes.keys():
            channel_counts[key] = 0

        while channel_pipes:
            for item in channel_pipes.items():
                if item[1].finished():
                    item[1].stop()
                    del channel_pipes[item[0]]
                    continue

                for address in item[1].read_all():
                    #for s in chat.app.channels.sub_list(ch):...
                    #if s[0] in dest_address or s[0]==chat.app.name_server.me.address:
                    #  continue
                    if address in dest_address:
                        continue
                    try:
                        ret_value = chat.node.call(address,message)
                        
                        #result = result + 'Message sent to '+ dest_names[i] +'.\n'
                        
                        #Type checking?
                        if type(ret_value) <> TupleType:
                            raise Error(_('unexpected value returned from call'))

                        dest_address.append(address)
                        #chcount = chcount + 1
                        channel_counts[item[0]] = channel_counts[item[0]]+1
                    except Error:
                        pass

            time.sleep(0.1)

            #dest_names.append(ch + ' (' + ("%d"%chcount) + ')')
            #result = 'Message sent to ' + english_list(dest_names) + '.\n' + prev_result
            #do_gtk(field.show,result+'Sending message...\n')

        for item in channel_counts.items():
            if item[1] == 1:
                result = result + (_('Message sent to %d person on channel %s.\n') % (item[1],item[0]))
            else:
                result = result + (_('Message sent to %d people on channel %s.\n') % (item[1],item[0]))
        do_gtk(field.show,result+_('Sending message...\n'))

        recall_list = [ ]
        for i in range(len(offline_list)):
            key   = random_bytes(name_bytes)
            lock  = hash.hash_of(key)
            crypt = offline_list[i].encrypt((key,message,long(time.time())))
            data  = {
                'type' : 'offline message',
                'crypt': crypt
            }
            name  = hash.hash_of('offline message '+offline_list[i].name)

            recall_list.append((offline_list[i].nickname,name,data,key))

            redundancy = chat.app.cache.publish([name],data,lock)

            if redundancy == cache_redundancy:
                str = offline_names[i] + _(' will get your message when they log on.\n')
            elif redundancy == 0:
                str = _('Could not store message for ') + offline_names[i] + '!\n'
            else:
                str = offline_names[i] + _(' will probably get your message when they log on:\n' +\
                    '  Message only stored with redundancy %d.\n')%redundancy

            result = result + str
            do_gtk(field.show,result + '\n')
        else:
            do_gtk(field.show,result + '\n')
            
        chat.recall_list = recall_list
    except Error, error:
        do_gtk(field.show,_('Error sending message:\n')+error.message+'\n')

    do_gtk(field.close)

def chat_receive_task(self, request,address, field1,quiet,send_time, add_to_history=1):
    field2 = field1.child()
    field3 = field2.child()

    #TODO: search for people not in acq list
    try:
        if not self.widget_visible:
            self.n_unseen_messages = self.n_unseen_messages + 1
            if self.n_unseen_messages == 1:
                title = _('1 unseen message')
            else:
                title = _('%d unseen messages') % self.n_unseen_messages
            self.app.set_title(title)
    
        has_attachment = (len(request) == 6)

        if type(request[1]) == type(()):
            sender_key_name = request[1][0]
            sender_name = request[1][1]
        else:
            # Used not to send username with key name
            sender_key_name = request[1]
            sender_name = 'unknown'

        try:
            sender = self.app.name_server.identify(sender_key_name,address)
        except Error:
            sender = None

        if sender:
            if has_attachment:
                sender.verify(request[3]+request[4],request[5])
            else:
                sender.verify(request[3],request[4])

        if sender:
            sender_nickname = sender.nickname
        else:
            sender_nickname = '<'+sender_name+'>'

        if has_attachment:
            attachment = safe_pickle.safe_loads(request[4])
            self.lock.acquire()
            i = 1
            while self.exec_vars.has_key("%s%d" % (sender_nickname,i)):
                i = i + 1
            self.exec_vars["%s%d"        % (sender_nickname,i)] = attachment
            self.exec_vars["%s%d_source" % (sender_nickname,i)] = address
            self.lock.release()
            if type(attachment) == type({}) and attachment.get('type',None) == 'file':
                field3_extra = _('\nContains attachment; type `!%s%d\' to see its details, `/take %s%d\' to retrieve.') % (sender_nickname,i,sender_nickname,i)
            else:
                field3_extra = _('\nContains attached python object; type `!%s%d\' to see.') % (sender_nickname,i)
        else:
            field3_extra = ''

        now_date  = time.strftime(_('%A %B %d, %Y'),time.localtime(time.time()))
        then_date = time.strftime(_('%A %B %d, %Y'),time.localtime(send_time))

        if now_date <> then_date:
            date_str = '\n'+then_date+' '
        else:
            date_str = '  '
        date_str = date_str + time.strftime(_('%I:%M %p'),time.localtime(send_time))
        str = '['+sender_nickname+' >>'

        do_gtk(field1.show,str+' ...]','blue')
        do_gtk(field2.show,' ' + request[3],'black')
        do_gtk(field3.show,date_str+field3_extra+'\n\n','light_grey') 

        if not quiet:
            do_gtk(beep)

        for item in request[2]:
            if item[0]==None and len(item)>=3 and item[2][0]=='#':
                str = str + ' '+item[2]
                self.app.channels.sub(item[2], 1)
                continue

            try:
                recipient = self.app.name_server.identify(item[0],item[1])

                recipient.lock.acquire()
                str = str + ' ' + recipient.nickname
                #self.refresh_user(recipient.nickname)
                recipient.lock.release()
            except Error,error :
                if len(item) >= 3:
                    str = str + ' <'+item[2]+'>'
                else:
                    str = str + ' ?'
            #Avoid excessive updating
            #do_gtk(field1.show,str+' ...]','blue')
        do_gtk(field1.show,str+']','blue')
    except Error:
        do_gtk(self.show,_('Bad message received from ')+repr(address)+'\n')

    do_gtk(field1.close)
    do_gtk(field2.close)
    do_gtk(field3.close)

    if add_to_history:
        self.lock.acquire()
        try:
            self.incoming_message_history.append((request,send_time))

            while len(self.incoming_message_history) > incoming_message_history_size:
                del self.incoming_message_history[0]
        finally:
            self.lock.release()
            
class Chat_field_text_group:
    def __init__(self, chat):
        self.chat = chat
        self.members = [ ]
        self.n_open = 0

    def add(self, member,after=None):
        if after:
            self.members.insert(self.members.index(after)+1,member)
        else:
            self.members.append(member)
        self.n_open = self.n_open + 1

    def closure(self):
        self.n_open = self.n_open - 1

        if not self.n_open:
            self.chat.text_hide_command_line()
            for member in self.members:
                sys.stdout.write(member.text)
            sys.stdout.flush()
            self.chat.text_show_command_line()
            self.chat = None

class Chat_field_text:
    def __init__(self,chat=None,group=None,after=None):
        if not group:
            group = Chat_field_text_group(chat)
        self.text = ''
        self.group = group
        group.add(self,after)
        
    def child(self):
        # Only function that can be called from other thread
        return Chat_field_text(group=self.group,after=self)

    def show(self,str,color='grey'):
        self.text = str

    def close(self):
        self.group.closure()

class Chat_field_gtk:
    def __init__(self,chat, start=None,end=None,index=None):
        chat.lock.acquire()
        try:
            if start:
                self.start = start
                self.end   = end
            else:
                self.start = chat.insertion_point
                self.end   = chat.insertion_point
            self.chat  = chat

            if index:
                chat.fields.insert(index,self)
            else:
                chat.fields.append(self)
        finally:
            chat.lock.release()

    def child(self):
        # Only function that can be called from other thread
        self.chat.lock.acquire()
        try:
            return Chat_field_gtk(self.chat,self.end,self.end,self.chat.fields.index(self)+1)
        finally:
            self.chat.lock.release()

    def show(self,str,color='grey'):
        import gtk
        import widgets

        self.chat.lock.acquire()
        try:
            if self.chat.widget.flags(gtk.DESTROYED):
                return

            pos = self.chat.widget.get_position()

            self.chat.widget.set_point(self.start)

            self.chat.widget.freeze()
            #self.chat.widget.insert_defaults(str)
            
            #if bold:
                #font = self.chat.bold_font
            #else:
                #font = self.chat.normal_font
            
            self.chat.widget.forward_delete(self.end-self.start)
            
            style = self.chat.widget.get_style()
            self.chat.widget.insert(
                #font,
                self.chat.font,
                #style.fg[STATE_NORMAL],
                getattr(widgets,color),
                style.base[gtk.STATE_NORMAL],
                str)
            
            self.chat.widget.thaw()

            change = len(str)-(self.end-self.start)
            
            pos = pos + change
            self.chat.horizon         = self.chat.horizon         + change
            self.chat.insertion_point = self.chat.insertion_point + change

            index = self.chat.fields.index(self)

            for item in self.chat.fields[index+1:]:
                item.start = item.start + change
                item.end   = item.end   + change
        
            self.chat.widget.set_position(pos)

            self.end = self.end + change
        finally:
            self.chat.lock.release()

    def close(self):
        self.chat.fields.remove(self)
        self.chat = None

class Chat(Task_manager):
    """ Base class of Chat_gtk and Chat_text, provides basic chat functions. """

    def __init__(self,app, use_gtk=1):
        Task_manager.__init__(self)
    
        self.node = app.node
        self.app  = app
        self.use_gtk = use_gtk

        # Visible status
        self.quiet       = 0
        self.activity    = ''

        # Detect repeated offline messages
        self.offline_message_buffer = [ ]

        # Store a copy of recent messages
        self.incoming_message_history = [ ]

        # While quiet, messages are buffered for later display
        self.unread_message_list = [ ]
        
        config = get_config("chat",{ })
        for item in ['quiet','activity','offline_message_buffer','incoming_message_history','unread_message_list']:
            if config.has_key(item):
                setattr(self,item,config[item])
        
        self.recall_list = [ ]

        # Detect repeated messages
        self.message_buffer = [ ]

        self.channel = [ ]

        self.history = [ ] # Commands
        self.history_position = 0

        # Results of /find
        self.search_list = [ ]

        # Python environment
        self.exec_vars = { 'app' : app, 'node' : app.node }

        # Command registry
        self._commands = { }  # name -> (n parameters, allow extra text, function)

        chat_commands.register(self)

    def repeated(self, address,call_id):
        if (address,call_id) in self.message_buffer:
            return 1
        self.message_buffer.append((address,call_id))
        if len(self.message_buffer) > 50:
            del self.message_buffer[0]
        return 0

    def start(self):
        Task_manager.start(self)
        
        self.app.name_server.set_status('chat',(not self.quiet,self.activity))
        
        self.set_normal_prompt()

        self.node.add_handler('chat message',self)
        self.node.add_handler('chat look',self)
        self.app.name_server.add_watch_callback(self)

        self.refresh_userlist()

        if self.app.gossip:
            field = self.get_field() 
            field.show(_('Fetching gossip...')+'\n\n\n\n\n\n')
            def on_gossiped(field=field,self=self,gossip=self.app.gossip):
                str = _("Today's hot gossip:\n")
                gossip.lock.acquire()
                n = 0
                for item in gossip.sorted_wodges()[:100]:
                    if not item[1].collapsed:
                        str = str + '   ' + item[1].wodge['subject'] + '\n'
                        n = n + 1
                        if n >= 5: 
                            break
                gossip.lock.release()

                if n == 0:
                    str = ''

                do_gtk(field.show,str)
                do_gtk(field.close)

            self.app.gossip.request_update(on_gossiped)

        self.retrieve_cached_messages()

        if self.quiet or self.activity <> '':
            self.show('\n')
            self.show_status()

    def stop(self):
        Task_manager.stop(self)
        self.node.remove_handler('chat message')
        self.node.remove_handler('chat look')
        self.app.name_server.remove_watch_callback(self)
        
        config = { }
        for item in ['quiet','activity','offline_message_buffer','incoming_message_history','unread_message_list']:
            config[item] = getattr(self,item)
        set_config("chat",config)

    def handle_watch(self, acq, because_of, **extras):
        acq.lock.acquire()
        try:
            nickname = acq.nickname
            watch    = acq.watch
        finally:
            acq.lock.release()

        if because_of == 'edit':
            self.lock.acquire()
            try:
                new_channel = [ ]
                for item in self.channel:
                    if item[:1] == '#' or \
                         self.app.name_server.nicknames.has_key(item):
                        new_channel.append(item)
            finally:
                self.lock.release()

            self.set_channel(new_channel)

        elif acq <> self.app.name_server.me:
            self.refresh_userlist()
            
        if not watch or (because_of <> 'connect' and because_of <> 'disconnect'):
            return

        self.lock.acquire()
        try:
            if not self.quiet:
                if because_of == 'connect':
                    action = _(' just arrived.\n')
                else:
                    if extras['message']:
                        action = _(' just left: ') + extras['message'] + '\n'
                    else:
                        action = _(' just left.\n')
                do_gtk(self.show,time.strftime(_('%I:%M %p')+' ',time.localtime(time.time())))
                do_gtk(self.show,nickname+action,'blue')
        finally:
            self.lock.release()

    def handle(self, request, address,call_id):
        self.lock.acquire()
        try:
            if request[0] == 'chat look':
                return (not self.quiet,self.activity)
            if request[0] == 'chat message':
                for recipient in request[2]:
                    if recipient[0] == self.app.name_server.public_key_name:
                        break
                    if len(recipient) >= 3 and recipient[2][:1] == '#' and \
                         self.app.channels.is_listening_to(recipient[2]) and \
                         request[1][0] <> self.app.name_server.public_key_name:
                        break
                else:
                    return None
            
                if not self.repeated(address,call_id):
                    if self.quiet:
                        self.unread_message_list.append((request,address,time.time()))
                        do_gtk(self.set_normal_prompt)
                    else:
                        Task(chat_receive_task,self,request,address, self.get_field(),self.quiet,time.time()).start()

                return (not self.quiet, self.activity)

        finally:
            self.lock.release()

    def refresh_userlist(self):
        pass

    def set_prompt(self, str):
        if not self.use_gtk:
            self.lock.acquire()
            try:
                if str == self.prompt:
                    return

                self.text_hide_command_line()
                self.prompt = str
                self.text_show_command_line()
            finally:
                self.lock.release()
            return

        # Kind of hacked...
        if self.quiet:
            color = 'green'
        else:
            color = 'blue'

        self.lock.acquire()
        try:
            insertion_point = self.insertion_point
            field = Chat_field_gtk(self, insertion_point,self.horizon)
            field.show(str,color)
            field.start = field.end
            field.show(' ','black')
            field.close()
            self.insertion_point = insertion_point
            self.horizon = field.end
        finally:
            self.lock.release()
        
    def set_normal_prompt(self):
        if self.shell:
            str = ''
        elif not self.use_gtk and self.command_text_lines:
            str = '>'
        else:
            if not self.channel:
                str = '['+self.app.config['name']+']'
            else:
                str = '['+self.app.config['name']+' >> '+string.join(self.channel,' ')+']'
                self.set_prompt(str)

            l = len(self.unread_message_list)
            if l:
                str = str + _(' (%d unread)')%l

        if not self.use_gtk:
            status = get_config('text_status',{ })
            status['unread'] = len(self.unread_message_list)
            set_config('text_status',status)

        self.set_prompt(str)

    def get_field(self):
        raise Error("define in subclass")
    
    def show(self, str,color='grey'):
        field = self.get_field()
        field.show(str,color)
        field.close()

    def retrieve_cached_messages(self):
        def task(self):
            field = self.get_field()
            do_gtk(field.show,_('Retrieving any messages sent while you were offline...\n'))
        
            pipe = self.node.retrieve(hash.hash_of('offline message '+self.app.name_server.public_key_name),cache_redundancy)
            list = [ ]
            while not pipe.finished():
                list.extend(pipe.read_all())
                time.sleep(1)
            pipe.stop()

            unique_messages = [ ]
            for item in list:
                if item[1] not in unique_messages:
                    unique_messages.append(item[1])

            message_list = [ ]
            for item in unique_messages:
                    if type(item) == type({}) and \
                         item.get('type','') == 'offline message' and \
                         item.has_key('crypt'):
                        try:
                            decrypt = self.app.name_server.decrypt(item['crypt'])

                            # Remove from caches
                            for thing in list:
                                if thing[1] == item:
                                    try:
                                        self.node.call(thing[0],('data cache remove',decrypt[0]))
                                    except Error:
                                        pass
                                    
                            message_list.append((decrypt[2],decrypt[1]))
                        except Error:
                            pass

            message_list.sort()
            self.lock.acquire()
            try:
                any = 0
                for item in message_list:
                    if item not in self.offline_message_buffer:
                        self.offline_message_buffer = [item] + self.offline_message_buffer[:50]
                        #Task(chat_receive_task,self,item[1],None, self.get_field(),1,item[0]).start()
                        self.unread_message_list.append((item[1],None,item[0]))
                        any = 1
                    else:
                        print _("Duplicate offline message.")
            finally:
                self.lock.release()

            if any:
                do_gtk(field.show,'\nYou received messages while you were offline. Type /read to read them.\n')
                do_gtk(self.set_normal_prompt)
            else:
                do_gtk(field.show,'\n')
            do_gtk(field.close)

        Task(task,self).start()

    def do_who(self):
        def task(self, acq, field, field_dots, n):
            acq.lock.acquire()
            name = acq.nickname
            acq.lock.release()

            acq.start_watching(self.node, 1)

            acq.lock.acquire()
            online = acq.online
            address = acq.address
            acq.lock.release()

            if online:
                do_gtk(field.show,name+' ')
                try:
                    result = self.node.call(address,('chat look',))
                    if not result[0]:
                        do_gtk(field.show,'*'+name+' ')
                except Error:
                    pass
            else:
                do_gtk(field.show,'')
            do_gtk(field.close)

            self.lock.acquire()
            n[0] = n[0] - 1
            if n[0] == 0:
                do_gtk(field_dots.show,'\n')
                do_gtk(field_dots.close)
            self.lock.release()

        field = self.get_field()
        field.show(_('Watched people currently online:\n'))
        new_field = field.child()
        field.close()
        field = new_field

        self.app.name_server.lock.acquire()
        values = self.app.name_server.acquaintances.values()
        self.app.name_server.lock.release()
        
        n = [0]
        fields = { }
        for item in values:
            if item.watch:
                n[0] = n[0]+1
                fields[item], field = field, field.child()

        if n[0]:
            field.show(' ...\n')
            for item in fields.keys():
                Task(task,self,item,fields[item],field,n).start()  
        else:
            field.show('\n')
            field.close()

    def show_status(self):
        if self.quiet:
            if self.activity == '':
                self.show(_('You are quiet.\n'))
            else:
                self.show(_('You are quiet: ')+self.activity+'\n')
        else:
            if self.activity == '':
                self.show(_('You are listening.\n'))
            else:
                self.show(_('What you are doing %s')%self.activity+'\n')

        if self.unread_message_list:
            self.show(_('You have unread messages. Type /read to read them.\n'))

    def set_channel(self, dests):
        self.lock.acquire()
        old_channel = self.channel
        self.channel = dests

        self.app.channels.check_join(dests)

        #for item in old_channel:
        #  self.refresh_user(item,0)
        #for item in self.channel:
        #  self.refresh_user(item,1)
        self.refresh_userlist()
        self.set_normal_prompt()
        self.lock.release()

    def set_status(self, quiet,activity):
        self.lock.acquire()
        self.quiet = quiet
        self.activity = activity
        self.lock.release()
        
        self.app.name_server.set_status('chat',(not quiet,activity))

    def send_message(self, dests, message, attachment=None):
        if self.quiet or self.activity <> '':
            self.show_status()
        Task(chat_tell_task,self,dests,message,self.get_field(),attachment).start()

    def register(self, name, min_param, max_param, function):
        """ Register a command. 
        
                A command takes a certain number of parameters separated by spaces,
                up to max_param, and not less than min_param. The last parameter may
                contain spaces.

                eg
                def mycommand(chat, param):
                    chat.show("You typed "+param+"\n")
                app.chat.register("mycommand",1,2,mycommand)
        """

        self.lock.acquire()
        self._commands[name] = (min_param, max_param, function)
        self.lock.release()
        
    def do(self, str):
        """ Exectute a chat command. """
        str = string.strip(str)

        #chat_commands.do(self, str)
        try:
	    if not str:
	        pass
		
            elif str[:2] == '//':
                self.run_shell(str[2:])

            elif str[0] == '/':
                list = string.split(str[1:],maxsplit=1)
                if len(list) < 1:
                    raise Error('')

                command = self._commands.get(list[0])
                if not command:
                    raise Error(_('I don\'t know how to do that.\nType /help for help on available commands.'))

                if len(list) == 2:
                    remainder = list[1]
                else:
                    remainder = ''

		param = string.split(remainder,maxsplit=command[1]-1)

                if len(param) < command[0]:
                    raise Error(_('/%s requires at least %d parameters.\nType /help for help on command usage.')%(list[0],command[0]))

                apply(command[2],(self,param))
            
            elif str[0] == '!':
                file = cStringIO.StringIO()
                stdout = sys.stdout
                stderr = sys.stderr
                sys.stdout = file
                sys.stderr = file
                try:
                    try:
                        result = eval(str[1:],self.exec_vars)
                        if result <> None:
                            print `result`
                    except SyntaxError:
                        exec str[1:]+'\n' in self.exec_vars
                except:
                    traceback.print_exc()
                sys.stdout = stdout
                sys.stderr = stderr
                self.show(file.getvalue())
                
            elif self.channel == [ ]:
                #self.show("Use /join to specify someone to talk to.\n")
                self.show("Before sending messages, you need to select a person or channel to talk to in the list to the right. If the person is not in the list, you need to search for them in the search box above and add them to your list of contacts.\n")

            else:
                self.send_message(self.channel[:],str)
        
        except Error, error:
            self.show(error.message+'\n')
        
        self.set_normal_prompt()
    
    def handle_user_entry(self, str):
        """ Handler user typing some text. """
        self.lock.acquire()
        try:
            if self.shell:
                self.set_prompt('')

                str = str + '\n'
                try:
                    while str:
                        n = os.write(self.shell_master,str)
                        str = str[n:]
                except OSError:
                    pass
                return
        finally:
            self.lock.release()

        #try:
        pos = string.find(str,'\n')
        if pos <> -1:
            self.history.append(str[:pos]+'\\')
        else:
            self.history.append(str)
        self.history_position = len(self.history)

        self.do(str)

        #finally: 
        #  self.set_normal_prompt()

class Chat_text(Chat):
    """ Text mode chat interface. """
    
    def __init__(self,app):
        Chat.__init__(self,app,0)

        self.prompt = ''
        self.command_text_lines = [ ]
        self.command_text = ''
        self.input_queue = ''
        self.escape_queue = [ ]
        self.command_visible = 0
        
        self.widget_visible = 1

        self.shell = None

    def stop(self):
        # main thread only!
        stdin_echo(0)
        self.text_hide_command_line()
        stdin_echo(1)

        Chat.stop(self)

    def run_shell(self, command):
        self.text_hide_command_line()
        stdin_echo(1)
        os.system(command)
        stdin_echo(0)
        self.text_show_command_line()

    def get_field(self):
        return Chat_field_text(self)

    def text_read_input(self):
        char = os.read(0,1)
        self.input_queue = self.input_queue + char

        if (char >= 'a' and char <= 'z') or \
             (char >= 'A' and char <= 'Z'):
            pos = string.find(self.input_queue,chr(27)+'[')
            if pos <> -1:
                self.escape_queue.append(self.input_queue[pos+2:])
                self.input_queue = self.input_queue[:pos]

    def text_get_escape(self,code):
        while 1:
            for i in range(len(self.escape_queue)):
                if self.escape_queue[i][-1] == code:
                    return self.escape_queue.pop(i)
            self.text_read_input()

    def text_backspace(self, n):
        while n:
            sys.stdout.write(chr(27)+'[6n')
            sys.stdout.flush()

            escape = self.text_get_escape('R')

            try:
                column = int(string.split(escape[:-1],';')[1]) - 1
            except:
                print _('Badly formatted escape sequence '),escape
                return

            if column == 0:
                sys.stdout.write(chr(27)+'[A'+chr(27)+'[999C'+chr(27)+'[K')
                sys.stdout.flush()
                n = n - 1
                continue
            
            if n < column:
                move = n
            else:
                move = column
            sys.stdout.write(chr(27)+'[%dD'%move+chr(27)+'[K')
            sys.stdout.flush()

            n = n - move

    def text_hide_command_line(self):
        #sys.stdout.write( (chr(8)+' '+chr(8))*(len(self.prompt)+len(self.command_text)) )
        #sys.stdout.flush()
        self.command_visible = self.command_visible - 1
        if self.command_visible == 0:
            self.text_backspace( len(self.prompt)+len(self.command_text)+1 )

    def text_show_command_line(self):
        self.command_visible = self.command_visible + 1
        if self.command_visible == 1:
            sys.stdout.write( self.prompt+' '+self.command_text )
            sys.stdout.flush()

    def text_advance_command_line(self):
        sys.stdout.write('\n')
        self.command_text = ''
        self.command_visible = 0

    #def text_mainloop(self):
    #  while 1:
    #    try:
    #      self.text_mainloop_inner()
#       return
#      except Error, error:
#        if error.message == 'sigint':
#         self.command_text = ''
#         self.command_text_lines = [ ]
#         self.set_normal_prompt()
#       else:
#         raise error

    def text_mainloop(self):
        self.command_text = ''

        set_config('text_status',{'running':1})

        stdin_echo(0)
        self.text_show_command_line()
        try:
            threadnice_mainloop_fake_start()
            while threadnice_mainloop_fake_is_running():
                threadnice_mainloop_fake_iteration()
                    
                if select.select([0],[ ],[ ],0.1)[0]:
                    self.text_read_input()

                while self.escape_queue:
                    item = self.escape_queue.pop(0)

                    if self.command_text_lines:
                        continue

                    if item == 'A' and self.history_position > 0:
                        self.history_position = self.history_position - 1
                        self.text_hide_command_line()
                        self.command_text = self.history[self.history_position]
                        self.text_show_command_line()

                    if item == 'B' and self.history_position < len(self.history):
                        self.history_position = self.history_position + 1
                        self.text_hide_command_line()
                        if self.history_position < len(self.history):
                            self.command_text = self.history[self.history_position]
                        else:
                            self.command_text = ''
                        self.text_show_command_line()

                while self.input_queue and self.input_queue <> chr(27) and self.input_queue[:2] <> chr(27)+'[':
                    char = self.input_queue[0]
                    self.input_queue = self.input_queue[1:]

                    if char == '\n':
                        line = self.command_text
                        self.text_advance_command_line()

                        if line[-1:] == '\\':
                            self.command_text_lines.append(line[:-1])
                            self.set_normal_prompt()
                            self.text_show_command_line()
                        else:
                            lines = self.command_text_lines
                            self.command_text_lines = [ ]
                            lines.append(line)
                            self.handle_user_entry(string.strip(string.join(lines,'\n')))
                            self.text_show_command_line()
                        continue

                    if char == chr(4):
                        self.text_hide_command_line()
                        self.show(_('Entering shell (Circle is still running)\n'))

                        self.lock.acquire()
                        old_quiet = self.quiet
                        old_activity = self.activity
                        self.lock.release()

                        if not old_quiet:
                            self.set_status(1,'(shell)')
        
                        command = 'exec $SHELL'
                        status = get_config('text_status',{ })
                        if status.get('directory',None):
                            command = 'cd '+quote_for_shell(status['directory'])+' ; '+command
                            check_directory = 0
                        else:
                            check_directory = 1

                        stdin_echo(1)
                        try:
                            self.do('//'+command)
                            self.show('\n')
                        finally:
                            stdin_echo(0)
                            self.text_show_command_line()
                        
                        if check_directory:
                            status = get_config('text_status',{ })
                            if not status.has_key('directory'):
                                self.show(_("Current working directory was not saved.\n"+\
                                                    "Suggest adding `circle -s` to your shell prompt (see /help).\n"))

                        if not old_quiet:
                            self.set_status(old_quiet,old_activity)

                        continue

                    if ord(char) < 32:
                        continue
                    
                    if char == chr(127):
                        if self.command_text:
                            self.command_text = self.command_text[:-1]
                            #sys.stdout.write(chr(8)+' '+chr(8))
                            #sys.stdout.flush()
                            self.text_backspace(1)
                    else:
                        sys.stdout.write(char)
                        sys.stdout.flush()
                        self.command_text = self.command_text + char

        finally:
            set_config('text_status',{'running':0})
            threadnice_mainloop_fake_stop()
            self.text_hide_command_line()
            stdin_echo(1)


class Chat_gtk(Chat):
    """ GTK chat interface. """

    def __init__(self,app):
        import gtk

        Chat.__init__(self,app,1)

        self.fields = [ ]
    
        self.widget = gtk.GtkText()
        self.widget.set_usize(10,10)
        self.widget.set_word_wrap(1)
        self.widget.set_editable(1)
        self.horizon = 0
        self.insertion_point = 0

        self.userlist_widget = gtk.GtkScrolledWindow()
        
        self.userlist_widget.ensure_style()
        font = self.userlist_widget.get_style().font
        
        self.userlist_widget.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)

        def on_event(_w,_e,self=self):
            length = self.widget.get_length() - self.horizon
            command = self.widget.get_chars(self.horizon,self.horizon+length)
            position = self.widget.get_position() - self.horizon
            
            if position < 0:
                return 1

            if _e.keyval == ord('c') and (_e.state & 5): # Ctrl-C
                self.lock.acquire()
                try:
                    if self.shell <> None:
                        os.kill(self.shell,2)
                except OSError:
                    pass
                self.lock.release()
                self.widget.emit_stop_by_name("key-press-event")
            
            if position == 0 and _e.keyval == 65361: # Left arrow
                self.widget.emit_stop_by_name("key-press-event")
            
            if string.find(command[:position],'\n') == -1 and \
                 _e.keyval == 65362: # Up arrow
                if self.history_position > 0:
                    self.history_position = self.history_position - 1
                    self.widget.set_position(self.horizon)
                    self.widget.freeze()
                    self.widget.set_point(self.horizon)
                    self.widget.forward_delete(length)
                    self.widget.insert_defaults(self.history[self.history_position])
                    self.widget.thaw()
                    self.widget.set_position(self.widget.get_length())
                self.widget.emit_stop_by_name("key-press-event")

            if string.find(command[position:],'\n') == -1 and \
                 _e.keyval == 65364: # Down arrow
                if self.history_position < len(self.history):
                    self.history_position = self.history_position + 1
                    if self.history_position < len(self.history):
                        str = self.history[self.history_position]
                    else:
                        str = ''
                    self.widget.set_position(self.horizon)
                    self.widget.freeze()
                    self.widget.set_point(self.horizon)
                    self.widget.forward_delete(length)
                    self.widget.insert_defaults(str)
                    self.widget.thaw()
                    self.widget.set_position(self.widget.get_length())
                self.widget.emit_stop_by_name("key-press-event")

            return 1
        self.widget.connect("key-press-event",on_event)

        def on_event(_w,_e,self=self):
            position = self.widget.get_position()
            if position < self.horizon:
                chars = self.widget.get_chars(position,self.horizon)
                if string.find(chars,'\n') == -1:
                    self.widget.set_position(self.horizon)
            return 1
        self.widget.connect_after("key-press-event",on_event)

        def on_delete(_w,start,end,_self=self):
            if start < _self.horizon:
                _self.widget.emit_stop_by_name("delete-text")
        self.widget.connect("delete-text",on_delete)

        def on_insert(_w,str,len,position, _self=self):
            import gtk
            position = _self.widget.get_position()
            if position < _self.horizon:
                _self.widget.emit_stop_by_name("insert-text")
                _self.widget.set_point(_self.widget.get_length())
                style = _self.widget.get_style()
                _self.widget.insert(
                    #_self.bold_font,
                    _self.font,
                    style.fg[gtk.STATE_NORMAL],
                    style.base[gtk.STATE_NORMAL],
                    str[:len])
                return
            if len == 1 and str[0] == '\n':
                if position <> _self.horizon and \
                     _self.widget.get_chars(position-1,position) == '\\':
                     return
                command = _self.widget.get_chars(_self.horizon,_self.widget.get_length())
                _self.widget.set_position(_self.widget.get_length())
                _self.horizon = _self.widget.get_position()
                _self.insertion_point = _self.horizon
                _self.show('\n')
                _self.widget.emit_stop_by_name("insert-text")

                command = string.replace(command,'\\\n','\n')
                _self.handle_user_entry(command)
                    
        self.widget.connect("insert-text",on_insert)

        self.widget_visible = 1
        self.n_unseen_messages = 0

        def on_draw(widget, event, self=self):
            import gtk
            self.widget_visible = (event.state <> gtk.GDK.VISIBILITY_FULLY_OBSCURED)
            if self.widget_visible:
                self.n_unseen_messages = 0
                self.app.set_title()
        self.widget.connect("visibility-notify-event",on_draw)
        self.widget.set_events(gtk.GDK.VISIBILITY_NOTIFY_MASK)

        # Shell command support
        self.shell = None   # (pid, write_pipe)

        try:
            openpty = None
            try:
                from pty import openpty
            except:
                from os import openpty
            self.shell_master, self.shell_slave = openpty()
            def callback(source, condition, fd=self.shell_master,self=self):
                try:
                    str = os.read(fd,256)
                except OSError:
                    str = ''
                str = string.replace(str,'\r','')
                if string.find(str,chr(7)) <> -1:
                    str = string.replace(str,chr(7),'')
                    beep()
                self.show(str)
            self.shell_tag = gtk.input_add(self.shell_master,5,callback)
        
            import termios

            # TERMIOS deprecated in 2.x series
            try:
                termios.ECHO
                TERMIOS = termios
            except:
                import TERMIOS

            attr = termios.tcgetattr(self.shell_slave)
            attr[3] = attr[3] & ~TERMIOS.ECHO
            termios.tcsetattr(self.shell_slave,TERMIOS.TCSANOW,attr)
        except:
            self.shell_master = None
            self.shell_slave  = None

    def start(self):
        import widgets

        self.font = self.widget.get_style().font

        self.blue_style   = self.widget.get_style().copy()
        for i in range(5):
            self.blue_style.fg[i] = widgets.blue
        self.green_style = self.widget.get_style().copy()
        for i in range(5):
            self.green_style.fg[i] = widgets.green

        Chat.start(self)

    def stop(self):
        import gtk

        self.lock.acquire()
        if self.shell_master <> None:
            gtk.input_remove(self.shell_tag)
            os.close(self.shell_master)
            os.close(self.shell_slave)

        try:
            if self.shell:
                os.kill(self.shell,2)
        except OSError:
            pass
        self.lock.release()

        Chat.stop(self)
 
    def run_shell(self, command):
        # only call from main thread

        if self.shell_master == None:
            do_gtk(self.show,_("No openpty in pty module.\n"))
            return

        try:
            pid = os.fork()
        except:
            event.set()
            do_gtk(self.show,_("Could not fork.\n"))

        if not pid:
            os.dup2(self.shell_slave, 0)
            os.dup2(self.shell_slave, 1)
            os.dup2(self.shell_slave, 2)
            os.execlp('/bin/sh','/bin/sh','-c',command)
            os._exit(0)
                
        self.lock.acquire()
        self.shell = pid
        self.lock.release()

        def timeout(self=self,pid=pid):
            result = os.waitpid(pid,os.WNOHANG)
            if result[0] == pid:
                self.lock.acquire()
                self.shell = None
                self.lock.release()
                #self.set_prompt('['+self.app.config['name']+']')
                self.set_normal_prompt()
                return 0
            return 1

        schedule_gtk(250, timeout)

    def refresh_userlist(self):
        def task(self):
            import gtk
            import widgets

            widget = gtk.GtkVBox(0,0)

            channels = self.app.channels.list.keys()
            channels.sort()
            for name in channels:
                try:
                    ch = self.app.channels.list[name]
                except ValueError:
                    continue

                hbox = gtk.GtkHBox(0,0)

                #mute = GtkCheckButton()
                mute = widgets.Toggle_icon_widget('<')
                mute.set_relief(gtk.RELIEF_NONE)
                def on_toggle(mute, name=name,self=self):
                    self.lock.acquire()
                    self.app.channels.mute(name,not mute.get_active())
                    do_gtk(self.widget.grab_focus)
                    self.lock.release()
                mute.connect("toggled", on_toggle)
                self.lock.acquire()
                mute.set_active(not ch['muted'])
                self.lock.release()
                hbox.pack_start(mute,0,0,0)
                widgets.tooltips.set_tip(mute,_("listen to channel %s")%name,"")
                
                #checker = GtkCheckButton()
                checker = widgets.Toggle_icon_widget('>')
                checker.set_relief(gtk.RELIEF_NONE)
                def on_toggle(checker, name=name,self=self):
                    self.lock.acquire()
                    if checker.get_active():
                        if name not in self.channel:
                            self.channel.append(name)
                            self.app.channels.mute(name,0)
                    else:
                        if name in self.channel:
                            self.channel.remove(name)
                    do_gtk(self.widget.grab_focus)
                    self.set_normal_prompt()
                    self.lock.release()
                checker.connect("toggled", on_toggle)
                self.lock.acquire()
                checker.set_active(name in self.channel)
                self.lock.release()
                hbox.pack_start(checker,0,0,0)
                widgets.tooltips.set_tip(checker,_("talk on %s")%name,"")

                button = widgets.Text_button_widget(name) 
                button.set_style(self.blue_style)
                button.set_relief(gtk.RELIEF_NONE)
                def on_click(button, name=name,self=self):
                    self.app.channels.edit_settings(name)
                button.connect("clicked",on_click)
                hbox.pack_start(button,1,1,0)
                widgets.tooltips.set_tip(button,_("view %s settings")%name)

                #label = GtkLabel(name)
                #label.set_alignment(0,0)
                #button.add(label)
         
                widget.pack_start(hbox,0,0,0)
 
            #widget.pack_start(GtkHSeparator(), FALSE,FALSE,0)

            self.app.name_server.lock.acquire()
            try:
                list = [ ]
                for item in self.app.name_server.nicknames.items():
                    if item[1] is not self.app.name_server.me:
                        if item[1].watch:
                            dist = item[1].distance
                            if dist == None: dist = 100.0
                        else:
                            dist = 101.0
                        if item[1].online:
                            top_sorter = 0
                        else:
                            top_sorter = 1
                        list.append(((top_sorter,dist),item[0],item[1]))
                list.sort()
            finally:
                self.app.name_server.lock.release()

            for item in list:
                item[2].lock.acquire()
                try:
                    name    = item[2].nickname
                    online  = item[2].online
                    status  = item[2].status.get('chat',(1,''))
                finally:
                    item[2].lock.release()

                addendum = ''
                if status[0] and status[1]:
                    addendum = '\n('+name+' '+status[1]+')'
                elif not status[0] and status[1]:
                    addendum = '\n('+name+_(' is quiet: ')+status[1]+')'
                
                hbox = gtk.GtkHBox(0,0)

                #checker = GtkCheckButton()
                checker = widgets.Toggle_icon_widget('>', 3)
                checker.set_relief(gtk.RELIEF_NONE)
                def on_toggle(checker, name=name,self=self):
                    self.lock.acquire()
                    if checker.get_active():
                        if name not in self.channel:
                            self.channel.append(name)
                    else:
                        if name in self.channel:
                            self.channel.remove(name)
                    do_gtk(self.widget.grab_focus)
                    self.set_normal_prompt()
                    self.lock.release()
                checker.connect("toggled", on_toggle)
                self.lock.acquire()
                checker.set_active(name in self.channel)
                self.lock.release()
                hbox.pack_start(checker,0,0,0)
                widgets.tooltips.set_tip(checker,_("talk to ")+name+addendum,"")
                    
                button = widgets.Text_button_widget(name) 
                button.set_relief(gtk.RELIEF_NONE)
                def on_click(button, name=name,self=self):
                    self.app.name_server.lock.acquire()
                    try:
                        if self.app.name_server.nicknames.has_key(name):
                            self.app.name_server.nicknames[name].edit(self.app.name_server)
                    finally:
                        self.app.name_server.lock.release()
                button.connect("clicked",on_click)
                hbox.pack_start(button,1,1,0)
                widgets.tooltips.set_tip(button,_("view %s's details")%name+addendum)
                         
                #label = GtkLabel(name)
                if online:
                    if status[0]:
                        button.set_style(self.blue_style)
                    else:
                        button.set_style(self.green_style)
                #label.set_alignment(0,0)
                #button.add(label)
                    
                widget.pack_start(hbox,0,0,0)
             

            for child in self.userlist_widget.children():
                self.userlist_widget.remove(child)

            self.userlist_widget.add_with_viewport(widget)
            widget.show_all()

        do_gtk(task,self)

    def get_field(self):
        return Chat_field_gtk(self)

