# Name server

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# TODO: make identify more efficient

# TODO: add check for errors on all try_address or identity_test s

# TODO: remove redundant passing of node

import time,os,types,struct,traceback,select,threading

import crypto.RSA
import crypto.rijndael
import crypto.number
RSA      = crypto.RSA
number   = crypto.number
rijndael = crypto.rijndael

from settings import *
from utility import *
from safe_pickle import *
from error import Error
import hash

if use_gtk:
    import GDK
    from gtk import *
    from widgets import *

def rand_from_time():
    str = struct.pack('d',time.time())
    value = 0
    for char in str:
        value = value ^ ord(char)
    return chr(value)
 
class Random_gtk:
    def __init__(self, transient_for=None):
        #self.file = open("/dev/random")

        self.buffer = ''
        #self.tag = input_add(self.file,1,self.callback)

        self.window = GtkWindow()
        self.window.set_title(_('Entropy collection'))
        self.window.set_modal(TRUE)
        if transient_for:
            self.window.set_transient_for(transient_for)
        self.window.set_border_width(10)
        self.window.connect("delete_event", lambda _w,_e: TRUE)
        self.window.connect("key-press-event", self.event)
        self.window.set_events(GDK.KEY_PRESS_MASK)

        vbox = GtkVBox(FALSE, 5)
        self.window.add(vbox)

        label = GtkLabel(_('Some random data is needed to generate a cryptographic identity for you.')+"\n")
        vbox.pack_start(label, FALSE,FALSE,0)
        
        self.progress_label = GtkLabel('')
        vbox.pack_start(self.progress_label, FALSE,FALSE,0)

        self.progress = GtkProgressBar('')
        vbox.pack_start(self.progress, FALSE,FALSE,0)

        self.window.show_all()

    #def callback(self, a,b):
    #  char = os.read(self.file.fileno(),1)
    #  self.buffer = self.buffer + char

    def event(self, widget,event):
        self.buffer = self.buffer + rand_from_time()

    def finish(self):
        #input_remove(self.tag)
        #self.file.close()
            
        self.progress_label.set_text(_('Done'))
        while events_pending():
            mainiteration()
        time.sleep(2)
        
        self.window.destroy()
        
    def randfunc(self, n):
        # Clear buffer so keypresses during calculation don't all arrive at
        # the same time
        while events_pending():
            mainiteration()
        self.buffer = ''
    
        self.progress_label.set_text(_('Please bang on the keyboard like a monkey.'))
        while len(self.buffer) < n:
            self.progress.update(float(len(self.buffer))/n)
            mainiteration()
        
        try:
            return self.buffer[:n]
        finally:
            #self.buffer = self.buffer[n:]
            self.progress_label.set_text(_('Calculating...'))
            self.progress.update(1.0)
            while events_pending():
                mainiteration()

class Random_text:
    def __init__(self, ignore=None):
        stdin_echo(0)
        print _('Some random data is needed to generate a cryptographic identity for you.')

    def finish(self):
        sys.stdout.write("\r"+_("Done.")+
                (" "*len(_("Please bang on the keyboard like a monkey."))))+"\n"
        sys.stdout.flush()
        stdin_echo(1)

    def randfunc(self, n):
        buffer = ''

        sys.stdout.write("\r"+_("Please bang on the keyboard like a monkey."))
        sys.stdout.flush()
        while select.select([0],[ ],[ ],0)[0]:
            os.read(0,1)

        while len(buffer) < n:
            os.read(0,1)
            buffer = buffer + rand_from_time()
        
        while select.select([0],[ ],[ ],0)[0]:
            os.read(0,1)

        sys.stdout.write("\r"+ _("Calculating...") +
                (" "*len(_("Please bang on the keyboard like a monkey."))))
        sys.stdout.flush()
        return buffer
            
if use_gtk:
    Random = Random_gtk
else:
    Random = Random_text

def key_name(key):
    return hash.hash_of(safe_dumps(key))

def identity_test(address, key, node):
    # Test identity
    tester = random_bytes(name_bytes)

    result = node.call(address,('identity test',tester))
    tester = hash.hash_of('identity test '+tester)

    try:
        if not crypto.RSA.construct(key).verify(tester,result):
            raise Error(_('Fake identity.'))
    except:
        traceback.print_exc()
        raise Error(_('Error in RSA.'))

def acquaintance_watch_task(acq,node):
    try:
        acq.lock.acquire()
        try:
            if not acq.running:
                return

            pipe = node.retrieve(acq.name, identity_redundancy)
        finally:
            acq.lock.release()

        try:
            failures = [ ]
            while not pipe.finished():
                list = pipe.read_all()
                for item in list:
                    # Todo: check it is a proper identity!!!!!

                    # Make sure we don't watch ourself (for startup detect other peer)
                    if item[1].get('peer-name',None) == node.name:
                        continue

                    if item[0] not in failures:
                        try:
                            acq.try_address(item[0],node)
                            if acq.online:
                                return
                        except Error:
                            failures.append(item[0])

                if not acq.running:
                    return

                # Ugh yuck bleah
                time.sleep(0.25)
        finally:
            pipe.stop()

        #acq.lock.acquire()
        #address = acq.address
        #acq.lock.release()
        #try:
        #  acq.try_address(address,node)
        #except Error:
        #  pass

    finally:
        acq.lock.acquire()
        acq.watched = 1
        acq.lock.release()

class Acquaintance(Task_manager):
    members = ['info','nickname','name','address','watch','distance']
    info_template = { 'type' : 'xxx' }

    def __init__(self, name_server):
        Task_manager.__init__(self)
        self.editor_open = 0
        
        self.watch     = 0
        self.remember  = 0
        self.distance  = None

        self.address   = None
        self.watching  = 0
        self.watched   = 0
        self.online    = 0
        self.status    = { }

        self.name_server = name_server
        
    def stop(self, node):
        if self.watching:
            self.stop_watching(node)

        Task_manager.stop(self, 1)

    def start_watching(self, node, wait=0):
        self.lock.acquire()
        try:
            if not self.running:
                return

            if not self.watching:
                self.watcher = {
                    'type' : 'watch',
                    'whom' : self.name
                }
                name = self.name
                watcher = self.watcher
                self.watching = 1
        
                self.lock.release()
                node.publish(hash.hash_of('watch '+name), watcher, identity_redundancy)
                self.lock.acquire()

                if not self.online:
                    self.watched  = 0
                    #Task(acquaintance_watch_task,self,node).start_no_overload()
                    Task(acquaintance_watch_task,self,node).start()
                else:
                    self.watched  = 1
        finally:
            self.lock.release()

        if wait:
            while 1:
                if self.watched:
                    break
                time.sleep(0.1)

    def stop_watching(self, node):
        self.lock.acquire()
        try:
            if not self.watching:
                return

            node.unpublish(self.watcher)
            self.watching = 0
            self.watched  = 1 # Wake up any waiters
        finally:
            self.lock.release()

    def sort_value(self):
        self.lock.acquire()
        try:
            if self.distance <> None:
                return self.distance
            elif self.watch:
                return 100.0
            else:
                return 101.0
        finally:
            self.lock.release()
            

    def try_address(self, address,node):
        self.lock.acquire()
        try:
            if self.online and self.address == address:
                return 0

            key = self.info['key']
        finally:
            self.lock.release()

        identity_test(address,key,node)

        self.info = node.call(address,('identity query',))
        check(self.info, Acquaintance.info_template)
        
        if self.info.get('peer-name') == node.name:
            return 0

        status = node.call(address,('identity watch',))

        if status == None:
            status = { }

        check(status, { })

        self.lock.acquire()
        try:
            was_online = self.online
            self.address  = address
            self.online   = 1
            self.watched  = 1
            self.status = status
            if self.info.has_key('up time'):
                self.connect_time = time.time() - self.info['up time']
            else:
                self.connect_time = time.time()
        finally:
            self.lock.release()

        self.start_watching(node)

        self.name_server.acquaintance_status_changed(self, 'discover')

        return not was_online

        #print 'found ', address, self.info

    def disconnection(self, address):
        self.lock.acquire()
        was_online = self.online
        if address == self.address:
            self.online = 0
        now_online = self.online
        self.lock.release()

        return was_online and not now_online

    def verify(self, str,sig):
        self.lock.acquire()
        try:
            if not crypto.RSA.construct(self.info['key']).verify(hash.hash_of(str),sig):
                raise Error(_('Bad signature.'))
        finally:
            self.lock.release()

    def encrypt(self, object):
        self.lock.acquire()
        try:
            key = random_bytes(16)

            pub_key   = crypto.RSA.construct(self.info['key'])
            encryptor = rijndael.rijndael(key, 16)
            
            encrypted_key  = pub_key.encrypt(key,'')[0]
            encrypted_text = encryptor.encrypt_cbc(safe_dumps(object))
            
            return (encrypted_key, encrypted_text)
        finally:
            self.lock.release()

    def to_map(self):
        self.lock.acquire()
        try:
            map = { }
            for key in self.members:
                map[key] = getattr(self,key)
            return map
        finally:
            self.lock.release()

    def from_map(self, map):
        self.lock.acquire()
        try:
            for key in map.keys():
                setattr(self,key,map[key])
            self.remember = 1
        finally:
            self.lock.release()

    def make_search_info(self):
        self.lock.acquire()
        try:
            return {
                'type': 'acquaintance',
                'nickname': self.nickname,
                'name': self.name,
                'data': self,
                'keywords': [ ]
            }
        finally:
            self.lock.release()
    
    def edit(self, ns):
        self.lock.acquire()
        try:
            if self.editor_open:
                return

            window = GtkWindow()
            window.set_border_width(10)
            window.set_title(_('Acquaintance: %s')%self.nickname)
            def on_destroy(_w, _self=self):
                _self.editor_open = 0
            window.connect("destroy",on_destroy)

            vbox = GtkVBox(FALSE,5)
            window.add(vbox)

            text = Text_widget(300)
            str = _('Name: ') + self.info.get('human-name','') + \
                        '\n' +_('Username: ') + self.info['name'] + \
                        '\n\n' + self.info.get('description','') + '\n'
            text.write(str, self.name_server.app)

            if self.info.has_key('timezone'):
                text.write(time.strftime(_('\nTheir current time is %I:%M %p'),time.gmtime(time.time()-self.info['timezone'])))

            if self.online:
                if self.info.has_key('up time'):
                    up_minutes = int((time.time()-self.connect_time)/60.0)
                    text.write(_('\nOnline for %d:%02d') % (int(up_minutes/60),up_minutes%60))
                def action(address=self.address,app=self.name_server.app,node=self.name_server.node):
                    import search
                    search.search_browse_files(address,node,app)
                def task(self, text=text,action=action,address=self.address):
                    name = hostname_of_ip(address[0])
                    do_gtk(text.write,_('\nCurrent address: '))
                    do_gtk(text.write_word,name + ':' + `address[1]`,
                                                    1,action)
                Task(task,self).start()
            elif self.watched:
                text.write(_('\nCurrently offline.'))
            else:
                text.write(_('\nCurrent status unknown.'))
                
            vbox.pack_start(text.widget, FALSE,FALSE,0)
            
            al = GtkAlignment(1,1,0,0)
            vbox.pack_start(al, FALSE,FALSE,0)
            signature = Signature_widget(self.name)
            al.add(signature.widget)

            al = GtkAlignment(0,0,0,0)
            vbox.pack_start(al, FALSE,FALSE,0)
            label = GtkLabel(_('Nickname:'))
            al.add(label)

            nick = GtkEntry()
            nick.set_text(self.nickname)
            vbox.pack_start(nick, FALSE,FALSE,0)
            
            distance = GtkCheckButton(_("Trust this person's gossip as:"))
            distance.set_active(self.distance <> None)
            vbox.pack_start(distance, FALSE,FALSE,0)

            if self.distance == None:
                initial_value = 1.0
            else:
                initial_value = self.distance
            distance_adj = GtkAdjustment(initial_value,0.0,20.5,0.5,0.5,0.5)

            distance_scale = GtkHScale(distance_adj)
            vbox.pack_start(distance_scale, FALSE,FALSE,0)

            distance_label = GtkLabel('')
            vbox.pack_start(distance_label, FALSE,FALSE,0)

            def value_changed(adjustment, label=distance_label):
                adjustment.set_value(int(adjustment.value*2+0.5)/2.0)
                label.set_text('')
                for item in distance_descriptions:
                    if adjustment.value >= item[0]:
                        label.set_text(item[1])
            def value_changed_2(adjustment, distance=distance):
                distance.set_active(TRUE)

            distance_adj.connect("value_changed",value_changed)
            distance_adj.connect("value_changed",value_changed_2)
            value_changed(distance_adj)

            if self.distance <> None:
                alignment = GtkAlignment(1,0,0,0)
                vbox.pack_start(alignment, FALSE, FALSE,0)

                hbox_gossip = GtkHBox(FALSE,5)
                alignment.add(hbox_gossip)

                gossip_1 = GtkButton(_(" View postings "))
                gossip_1.connect("clicked",lambda button, self=self,ns=ns: ns.app.gossip.show_gossip(person=self.name,update=1))
                hbox_gossip.pack_start(gossip_1, FALSE,FALSE,0)
                gossip_2 = GtkButton(_(" View their perspective "))
                gossip_2.connect("clicked",lambda button, self=self,ns=ns: ns.app.gossip.show_gossip(perspective=self.name,update=1))
                hbox_gossip.pack_start(gossip_2, FALSE,FALSE,0)

                def distance_changed(ignore=None, distance=distance,gossip_1=gossip_1,gossip_2=gossip_2):
                    active = distance.get_active()
                    gossip_1.set_sensitive(active)
                    gossip_2.set_sensitive(active)
                distance.connect("toggled",distance_changed)
                distance_changed()
            
            vbox.pack_start(GtkHSeparator(), FALSE,FALSE,0)

            watch = GtkCheckButton(_('Tell me when this person logs in and out'))
            watch.set_active(self.watch)
            vbox.pack_start(watch, FALSE,FALSE,0)
            remember = GtkCheckButton(_('Remember this person between sessions'))
            remember.set_active(self.remember)
            vbox.pack_start(remember, FALSE,FALSE,0)

            hbox = GtkHBox(FALSE,5)
            vbox.pack_end(hbox, FALSE,FALSE,0)

            button = GtkButton(_("OK"))
            def on_ok(_b, _self=self,_ns=ns,_nick=nick,_window=window,_watch=watch,_remember=remember,_distance=distance,_distance_adj=distance_adj):
                try:
                    _ns.change_nickname(_self,_nick.get_text())
                except Error, error:
                    error.show(_window)
                    return
                _self.lock.acquire()
                try:
                    _self.watch    = _watch.get_active()
                    _self.remember = _remember.get_active()

                    if _distance.get_active():
                        _self.distance = _distance_adj.value
                    else:
                        _self.distance = None
                finally: 
                    _self.lock.release()

                if _watch.get_active():
                    _self.start_watching(_ns.node)

                _window.destroy()

                # Just in case we crash before stopping
                _ns.save()

                _ns.acquaintance_status_changed(_self, 'edit')

            button.connect("clicked",on_ok)
            hbox.pack_start(button, TRUE,TRUE,0)
            
            button = GtkButton(_("Cancel"))
            button.connect("clicked",lambda _b,_window=window:_window.destroy())
            hbox.pack_start(button, TRUE,TRUE,0)
            
            def on_changed(_w, _remember=remember):
                _remember.set_active(TRUE)
            nick.connect("changed",on_changed)
            distance.connect("toggled",lambda widget, on_changed=on_changed : widget.get_active() and on_changed(widget))
            distance_adj.connect("value_changed",on_changed)

            show_window(window, 'acquaintance')
            self.editor_open = 1
        finally:
            self.lock.release()

def name_server_add_watcher_task(ns,address):
    ns.lock.acquire()
    try:
        if not ns.running:
            return

        if address in ns.watchers:
            return

        ns.watchers.append(address)
    finally:
        ns.lock.release()

    try:
        ns.node.call(address,('identity connecting',ns.public_key_name))
    except Error:
        pass

def name_server_watch_poller(self):
    pipe = self.node.retrieve(hash.hash_of('watch '+self.public_key_name), identity_redundancy)
    watcher = {
        'type' : 'watch',
        'whom' : self.public_key_name
    }
        
    try:
        while not pipe.finished():
            for item in pipe.read_all():
                if item[1] == watcher:
                    self.lock.acquire()
                    try: 
                        if item[0] not in self.watchers:
                            Task(name_server_add_watcher_task,self,item[0]).start_no_overload()
                    finally:
                        self.lock.release()

            if not self.running:
                return
            # Ugh yuck bleah
            time.sleep(2)
    finally:
        pipe.stop()

def refresh_offline_presence_task(self):
    pipe = self.node.retrieve(self.public_key_name_offline, cache_redundancy)

    list = [ ]
    try:
        while not pipe.finished():
            for item in pipe.read_all():
                if type(item[1]) == types.DictType and \
                     item[1].get('type') == 'identity offline' and \
                     item[1].get('salt'):
                    list.append(item)
            time.sleep(2)
            if not self.running:
                return
    finally:
        pipe.stop()

    if len(list) > 2:
        print _("%d peers holding your offline presence.") % len(list)

    for item in list:
        address, value = item

        key = hash.hash_of(safe_dumps(self.sign(value['salt'])))

        try:
            self.node.call(address, ('data cache remove',key)) 
	except Error:
	    pass

    self.lock.acquire()
    try:
        package = {
            'name'       : self.info['name'],
            'human-name' : self.info['human-name'],
            'description': self.info['description'],
            'timezone'   : self.info['timezone'],
            'key'        : self.public_key,
            'keywords'   : self.info['keywords'],
        }
    finally:
        self.lock.release()

    package_dumped = safe_dumps(package)
    signature = self.sign(package_dumped)

    value = {
        'type' : 'identity offline',
        'package' : package_dumped,
        'signature' : signature,
        'salt' : random_bytes(name_bytes)
    }

    lock = hash.hash_of(hash.hash_of(safe_dumps(self.sign(value['salt']))))
    
    publications = [ self.public_key_name_offline, self.service_name ]
    for item in package['keywords']:
        publications.append(hash.hash_of('identity-name '+item))

    self.app.cache.publish(publications,value,lock, 2)

class Name_server(Task_manager):
    def __init__(self, app,node, transient_for=None):
        Task_manager.__init__(self)

        self.app = app
        self.node = node

        self.key = get_config("private_key",None)

        if self.key == None:
            random = Random(transient_for)
            self.key = crypto.RSA.deconstruct(crypto.RSA.generate(1024,random.randfunc))
            random.finish()
            set_config("private_key",self.key)

        self.key  = crypto.RSA.construct(self.key)
        self.public_key = crypto.RSA.deconstruct(self.key.publickey())
        self.public_key_name = key_name(self.public_key)
        self.public_key_name_offline = hash.hash_of('identity-offline '+key_name(self.public_key))
        self.name   = None
        self.info   = { }
        self.status = { }    # eg chat status stored under 'chat' key

        # For searching for all people
        self.service_name = hash.hash_of("service identity")

        self.acquaintances = { }
        self.nicknames     = { }

        self.watchers      = [ ]

        self.watch_callbacks = [ ]

        self.start_time = time.time()

        self.aborted = 0

    def add_watch_callback(self, object):
        self.lock.acquire()
        self.watch_callbacks.append(object)
        self.lock.release()
    
    def remove_watch_callback(self, object):
        self.lock.acquire()
        self.watch_callbacks.remove(object)
        self.lock.release()

    def set_status(self, key,value):
        self.lock.acquire()
        try:
            if not self.status.has_key(key) or \
                 self.status[key] <> value:
                self.status[key] = value

                def task(self, address, status):
                    try:
                        self.node.call(address,('identity status changed',self.public_key_name,status))
                    except Error:
                        pass
                for item in self.watchers:
                    Task(task,self,item,self.status).start()
        finally:
            self.lock.release()
    
    def set_name(self, name, human_name, description):
        if name == self.info.get('name') and \
             human_name == self.info.get('human-name') and \
             description == self.info.get('description'):
            return
    
        now = time.time()
        timezone = long(time.mktime(time.gmtime(now)) - now)

        new_info = {
            'type'       : 'identity',
            'name'       : name,
            'human-name' : human_name,
            'description': description,
            'timezone'   : timezone,
            'key'        : self.public_key,
            'peer-name'  : self.node.name, # Used in startup
            'keywords'   : [ name ] + string.split(string.lower(human_name))
        }

        self.lock.acquire()
        running = self.running
        self.lock.release()

        if running:
            self.stop('Reconfiguring')

        self.name = name
        self.info = new_info

        if running:
            self.start()

    def get_info(self):
        self.info['up time'] = int(time.time() - self.start_time) 
        return self.info
    
    def start(self, status_monitor=None):
        Task_manager.start(self)

        acq_list = get_config('acquaintances',[ ])
        for item in acq_list:
            try:
                check(item['info'], Acquaintance.info_template)
            except Error:
                continue

            acq = Acquaintance(self)
            acq.from_map(item)
            self.acquaintances[acq.name] = acq
            self.nicknames[acq.nickname] = acq
            acq.start()
            
        self.me = self.make_acquaintance(self.info)

        # Other me may want to test identity
        # May start chatting before test complete
        self.node.add_handler('identity test', self)
        self.node.add_handler('identity query', self)
        self.node.add_handler('identity watch', self)
        self.node.add_handler('identity connecting', self)
        self.node.add_handler('identity status changed', self)
        self.node.add_handler('identity disconnecting', self)
        self.node.add_handler('identity abort', self)

        self.node.publish(self.public_key_name,self.get_info, identity_redundancy)
        self.node.publish(self.public_key_name_offline,self.get_info, identity_redundancy)
        self.node.publish(self.service_name,self.get_info, identity_redundancy)
        for item in self.info['keywords']:
            self.node.publish(hash.hash_of('identity-name '+item),self.get_info, identity_redundancy)
 
     
        def startup_task(self, status_monitor=status_monitor):
            list = self.acquaintances.values()
            list.sort(lambda x,y: cmp(x.sort_value(),y.sort_value()))
            for item in list:
                item.start_watching(self.node)

                #start watching tends to breed, try to make sure we don't get
                #too many threads.
                #yes, this is hacky
                #print item.nickname, threading.activeCount()
                #time.sleep(0.25)

                while 1:
                    time.sleep(0.25)
                    if threading.activeCount() < 40:
                        break
                 
            self.me.start_watching(self.node,1)
            self.me.lock.acquire()
            online = self.me.online
            address = self.me.address
            self.me.lock.release()
            if online:
                if status_monitor:
                    status_monitor(_('Shutting down your other peer.'))
                while 1:
                    try:
                        self.node.call(address, ('identity abort',))
                    except Error:
                        break
                    time.sleep(4)
            
            self.me.lock.acquire()
            self.me.online  = 1
            self.me.address = self.node.address
            self.me.connect_time = time.time()
            self.me.lock.release()

            # Task to retrieve existing watchers
            # Task to poll existing watchers
            #Task(name_server_watch_poller_task,self).start()
            self.start_poller(0, 10*60,name_server_watch_poller,self)

            refresh_offline_presence_task(self)
            
        Task(startup_task,self).start()

    def stop(self, quit_message):
        self.save()

        # Inform watchers of going offline
        self.lock.acquire()
        try:
            self.running = 0
        
            def task(self, address, quit_message=quit_message):
                try:
                    self.node.call(address,('identity disconnecting',self.public_key_name,quit_message))
                except Error:
                    pass

            if not self.aborted:
                for item in self.watchers:
                    Task(task,self,item).start()
            
            self.watchers      = [ ]
            
            list = self.acquaintances.values()
        finally:
            self.lock.release()
            
        for item in list:
            def task(self,item):
                item.stop(self.node)
            Task(task,self,item).start()
    
        Task_manager.stop(self)

        self.node.unpublish(self.get_info)

        self.node.remove_handler('identity test')
        self.node.remove_handler('identity query')
        self.node.remove_handler('identity watch')
        self.node.remove_handler('identity connecting')
        self.node.remove_handler('identity status changed')
        self.node.remove_handler('identity disconnecting')
        self.node.remove_handler('identity abort')
        
        # Remove cyclical dependancy in ref counting
        self.acquaintances = { }
        self.nicknames     = { }

    def save(self):
        self.lock.acquire()
        
        list = [ ]
        for item in self.acquaintances.values():
            if item.remember:
                list.append(item.to_map())
        set_config('acquaintances',list)

        self.lock.release()

    def sign(self, str):
        return self.key.sign(hash.hash_of(str),'')

    def decrypt(self, crypt):
        # Todo: catch errors
        self.lock.acquire()
        try:
            encrypted_key, encrypted_text = crypt

            key = self.key.decrypt((encrypted_key,))
            while len(key) < 16:
                key = chr(0)+key

            decryptor = rijndael.rijndael(key, 16)
            text      = decryptor.decrypt_cbc(encrypted_text)

            return safe_loads(text)
        finally:
            self.lock.release()

    def acquaintance_status_changed(self, acq, because_of, **extras):
        self.lock.acquire()
        callbacks = self.watch_callbacks[:]
        self.lock.release()

        for item in callbacks:
            apply(item.handle_watch,(acq,because_of),extras)

    def handle(self, request, address,call_id):
        if request[0] == 'identity test':
            check_is_name(request[1])
            return self.sign('identity test ' + request[1])
        elif request[0] == 'identity query':
            return self.get_info()
        elif request[0] == 'identity watch':
            self.lock.acquire()
            if address not in self.watchers:
                self.watchers.append(address)
            status = self.status
            self.lock.release()
            return status
        elif request[0] == 'identity abort':
            def abort_task(self, address=address):
                try:
                    identity_test(address,self.public_key,self.node)
                except Error:
                    return
                self.aborted = 1
                do_gtk(self.app.shutdown,_('Reconnecting from different machine'))

            Task(abort_task,self).start()
            return None
        elif request[0] == 'identity connecting':
            self.lock.acquire()
            if self.acquaintances.has_key(request[1]):
                acq = self.acquaintances[request[1]]
                self.lock.release()

                #if len(request) > 2:
                #  status = request[2]
                #else:
                #  status = { }

                def task(ns,acq,address,call_id,node):
                    try:
                        if not acq.try_address(address,node):
                            return
                    except Error:
                        return

                    #acq.lock.acquire()
                    #acq.status = status
                    #acq.lock.release()
                    ns.acquaintance_status_changed(acq, 'connect')
                Task(task,self,acq,address,call_id,self.node).start()

                return None
            else:
                self.lock.release()
                return Error(_("I don't know you."))
        elif request[0] == 'identity status changed':
            self.lock.acquire()
            try:
                acq = self.acquaintances.get(request[1],None)
                if acq:
                    acq.lock.acquire()
                    try:
                        acq.status = request[2]
                    finally:
                        acq.lock.release()
                    self.acquaintance_status_changed(acq, 'status changed')
            finally:
                self.lock.release()
        elif request[0] == 'identity disconnecting':
            self.lock.acquire()
            if address in self.watchers:
                self.watchers.remove(address)
            
            if self.acquaintances.has_key(request[1]):
                acq = self.acquaintances[request[1]]
                self.lock.release()
                if acq.disconnection(address):
                    self.acquaintance_status_changed(acq, 'disconnect', message=request[2])
                return None
            else:
                self.lock.release()
                return Error(_("I don't know you."))

    def choose_nickname(self,name):
        if name == '':
            name = 'unknown'
        i = 0
        nick = name
        while self.nicknames.has_key(nick):
            i = i + 1
            nick = name + ('%d'%i)
        return nick

    def change_nickname(self, acq,new_nick):
        self.lock.acquire()
        try:
            if new_nick == acq.nickname:
                return
            if self.nicknames.has_key(new_nick):
                raise Error(_("There is someone else with the nickname %s")%new_nick)
            del self.nicknames[acq.nickname]
            self.nicknames[new_nick] = acq
            acq.nickname = new_nick
        finally:
            self.lock.release()
            

    def locate(self, nick):
        if not self.nicknames.has_key(nick):
            raise Error(_('That person is not in the list of acquaintances.\n'+\
                                        'You need to search for %s before you can talk to them.')
                                    %nick)
        return self.nicknames[nick]

    def edit_acquaintance(self, info,address):
        def task(ns, info,address):
            status = Status()
            do_gtk(status.show,_("Authenticating"),_("Testing the person's identity."))
            try:
                acq = ns.make_acquaintance(info,address)
            except Error, error:
                do_gtk(error.show)
                do_gtk(status.close)
                return
            
            do_gtk(status.close)
            do_gtk(acq.edit,ns)
        Task(task,self,info,address).start()

    def identify(self, name,address):
        self.lock.acquire()
        if self.acquaintances.has_key(name):
            acq = self.acquaintances[name]
            self.lock.release()

            if address:
                acq.try_address(address,self.node)
                
            return acq

        self.lock.release()

        if not address:
            raise Error(_('no address'))

        info = self.node.call(address, ('identity query',))
        return self.make_acquaintance(info,address)
 
    def make_acquaintance(self, info,address=None):
        if not info.has_key('type') or \
             not info.has_key('name') or \
             not info.has_key('key')  or \
             info['type'] not in ['identity','identity offline demangled']:
            raise Error(_('Not an identity.'))

        check_type(info['name'],types.StringType)
        check_is_public_key(info['key'])
            
        # done in try_address
        #identity_test(address,info['key'],self.node)

        name = key_name(info['key'])
            
        self.lock.acquire()
        if self.acquaintances.has_key(name):
            self.lock.release()
            if address:
                self.acquaintances[name].try_address(address,self.node)
            return self.acquaintances[name]

        acq = Acquaintance(self)
        acq.info = info
        acq.name = name
        acq.nickname = self.choose_nickname(info['name'])
        self.acquaintances[name]     = acq
        self.nicknames[acq.nickname] = acq
        self.lock.release()

        acq.start()

        if address:
            acq.try_address(address,self.node)
        
        acq.start_watching(self.node,0)
        self.acquaintance_status_changed(acq, "create")
        
        return acq


        #self.lock.acquire()
        #try:
        #  if self.acquaintances.has_key(name):
        #    return self.acquaintances[name]
#
#      acq = Acquaintance()
#      acq.info = info
#      acq.address = address
#      acq.name = name
#      acq.nickname = self.choose_nickname(info['name'])
#
#      self.acquaintances[name] = acq
#      self.nicknames[acq.nickname] = acq
#
#      acq.start()
#      return acq
#    finally:
#      self.lock.release()


            
